/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	Vector functions.											*/
/*	Matrix manipulation functions.								*/
/*	Written by Stephen Chenney. Dec 1993.						*/
/*																*/
/****************************************************************/
/*																*/
/*																*/
/****************************************************************/

#include <stdio.h> /* for the NULL definition */
#include <math.h>
#include "Vector.h"


/* A small number for floating point comparisons. */
/* A floating point == 0.0 function. */
#define iszero(f)	(((f) < 1.e-12) && ((f) > -1.e-12))


int
VRead(char *s, Vector* v)
/****************************************************************/
/*																*/
/*	Reads a vector from a file. (may be stdin).					*/
/*	Returns 0 on failure, 1 otherwise.							*/
/*																*/
/****************************************************************/
{
	if ((sscanf(s, "%lf %lf %lf", &((*v).x), &((*v).y), &((*v).z))) != 3)
		return 0;
	else
		return 1;
}



/****************************************************************/
/*																*/
/*	USAGE INFORMATION:											*/
/*	These functions return a Vector*, which is the same as		*/
/*	 a Matrix.													*/
/*	They take as a final argument the thing you want the		*/
/*	 answer put in.  This is actually what is returned.			*/
/*	 The comments at the start of the vector.c file apply		*/
/*	 equally well here.											*/
/*																*/
/****************************************************************/


Vector*
MMMul(Matrix m1, Matrix m2, Matrix r)
/****************************************************************/
/*																*/
/*	Multiplies 2 matrices: r = m1m2.							*/
/*																*/
/****************************************************************/
{
	/* This is ugly code, caused by the vector structure. */

	Vector col1, col2, col3;  /* Columns of m2. */
	Vector row1, row2, row3;  /* Rows of m1. */

	/* Get the columns of m2, then it's easy. */
	col1.x = m2[0].x; col1.y = m2[1].x; col1.z = m2[2].x;
	col2.x = m2[0].y; col2.y = m2[1].y; col2.z = m2[2].y;
	col3.x = m2[0].z; col3.y = m2[1].z; col3.z = m2[2].z;
	/* Need copies of the rows of m1 because they may be written over. */
	row1 = m1[0]; row2 = m1[1] ; row3 = m1[2];

	r[0].x = VDot(row1, col1);
	r[0].y = VDot(row1, col2);
	r[0].z = VDot(row1, col3);
	r[1].x = VDot(row2, col1);
	r[1].y = VDot(row2, col2);
	r[1].z = VDot(row2, col3);
	r[2].x = VDot(row3, col1);
	r[2].y = VDot(row3, col2);
	r[2].z = VDot(row3, col3);

	return r;
}


Vector*
MInvert(Matrix m, Matrix r)
/****************************************************************/
/*																*/
/*	Inverts the matrix m and puts it in r.						*/
/*	If the matrix m is singular, returns NULL and leaves r		*/
/*	 unchanged.  Else returns r.								*/
/*																*/
/****************************************************************/
{
	/* This does it by the "formula" method.*/
	/* a[i][j] = det(adjoint[i][j])/det(b)	*/

	double temp;	/* for the determinant. */
	double a11, a22, a33; /* Various adjoint determinants. */

	a11 = m[1].y * m[2].z - m[1].z * m[2].y;
	a22 = m[0].y * m[2].z - m[0].z * m[2].y;
	a33 = m[0].y * m[1].z - m[0].z * m[1].y;
	temp =	m[0].x * a11 - m[1].x * a22 + m[2].x * a33;
	
	/* Test for singularity. */
	if (iszero(temp)) return NULL;

	r[0].x = a11 / temp;
	r[0].y = -a22 / temp;
	r[0].z = a33 / temp;
	r[1].x = (m[1].z * m[2].x - m[1].x * m[2].z) / temp;
	r[1].y = (m[0].x * m[2].z - m[0].z * m[2].x) / temp;
	r[1].z = (m[0].z * m[1].x - m[0].x * m[1].z) / temp;
	r[2].x = (m[1].x * m[2].y - m[1].y * m[2].x) / temp;
	r[2].y = (m[0].y * m[2].x - m[0].x * m[2].y) / temp;
	r[2].z = (m[0].x * m[1].y - m[0].y * m[1].x) / temp;

	return r;
}



