/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**	View.c : C file for the View widget class.
**
**	Created: 19/03/94
*/

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "ViewP.h"

static XtResource resources[] = {
#define offset(field) XtOffsetOf(ViewRec, view.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
	{ XtNexposeCallback, XtCCallback, XtRCallback, sizeof(XtCallbackList),
	  offset(expose_callback), XtRCallback, NULL },
	{ XtNresizeCallback, XtCCallback, XtRCallback, sizeof(XtCallbackList),
	  offset(resize_callback), XtRCallback, NULL },
	{ XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	  offset(foreground_pixel), XtRString, XtDefaultForeground }
#undef offset
};


/* Prototypes for various procedures. */
static void Resize(Widget w);
static void Redisplay(Widget w, XEvent *event, Region region);


ViewClassRec viewClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"View",
    /* widget_size		*/	sizeof(ViewRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	NULL,
    /* initialize_hook		*/	NULL,
    /* realize			*/	XtInheritRealize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	NULL,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* view fields */
    /* empty			*/	0
  }
};

WidgetClass viewWidgetClass = (WidgetClass)&viewClassRec;


/*	void Resize(Widget w)
**	Returns if the widget is unrealized.
**	Otherwise, clears the window with exposures TRUE.
*/
static void
Resize(Widget w)
{
	XPoint	new_size;

	if (XtIsRealized(w))
	{
		new_size.x = w->core.width;
		new_size.y = w->core.height;
		XtCallCallbacks(w, XtNresizeCallback, (XtPointer)&new_size);
	}
}


/*	void Redisplay(Widget w, XEvent *event, Region region);
**	The expose procedure.
**	Calls the Callback for expose.
*/
static void
Redisplay(Widget w, XEvent *event, Region region)
{
	XtCallCallbacks(w, XtNexposeCallback, (XtPointer)&region);
}

