/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**	genray : a generic raytracer.
**
**	extents.c : extent manipulation functions.
*/

#include "genray.h"

void Compound_Combine(Instance *obj, Vector *points);
void Square_Combine(Instance *obj, Vector *points);
void Cube_Combine(Instance *obj, Vector *points);
void Pyramid_Combine(Instance *obj, Vector *points);
void Torus_Combine(Instance *obj, Vector *points);

int	i;

void
Compound_Extent(Instance *obj, Vector* w_to_v, Vector disp)
/****************************************************************/
/*																*/
/*	Determines the extents for a compound object.				*/
/*																*/
/****************************************************************/
{
	Vector	obj1_points[8];
	Vector	obj2_points[8];
	Vector	view_points[16];
	Vector	min_space, max_space;
	Vector	temp_v;
	double	temp_d;

	/* Get the corners for each child. */
	switch (obj->object1->type)
	{
		case plane :
			/* The person specifying such an object is crazy. */
			/* The projection and spacial extent for the	*/
			/*	 whole compound object will be the whole	*/
			/*	 of space.									*/
			VNew(HUGE, HUGE, HUGE, obj1_points[0]);
			VNew(HUGE, HUGE, -HUGE, obj1_points[1]);
			VNew(HUGE, -HUGE, HUGE, obj1_points[2]);
			VNew(HUGE, -HUGE, -HUGE, obj1_points[3]);
			VNew(-HUGE, HUGE, HUGE, obj1_points[4]);
			VNew(-HUGE, HUGE, -HUGE, obj1_points[5]);
			VNew(-HUGE, -HUGE, HUGE, obj1_points[6]);
			VNew(-HUGE, -HUGE, -HUGE, obj1_points[7]);
			break;
		case square :
			Square_Combine(obj->object1, obj1_points);
			break;
		case sphere:
		case cube:
		case cylinder:
			Cube_Combine(obj->object1, obj1_points);
			break;
		case cone:
			Pyramid_Combine(obj->object1, obj1_points);
			break;
		case torus:
			Torus_Combine(obj->object1, obj1_points);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			Compound_Combine(obj->object1, obj1_points);
			break;
	}

	switch (obj->object2->type)
	{
		case plane :
			/* The person specifying such an object is crazy. */
			/* The projection and spacial extent for the	*/
			/*	 whole compound object will be the whole	*/
			/*	 of space.									*/
			VNew(HUGE, HUGE, HUGE, obj2_points[0]);
			VNew(HUGE, HUGE, -HUGE, obj2_points[1]);
			VNew(HUGE, -HUGE, HUGE, obj2_points[2]);
			VNew(HUGE, -HUGE, -HUGE, obj2_points[3]);
			VNew(-HUGE, HUGE, HUGE, obj2_points[4]);
			VNew(-HUGE, HUGE, -HUGE, obj2_points[5]);
			VNew(-HUGE, -HUGE, HUGE, obj2_points[6]);
			VNew(-HUGE, -HUGE, -HUGE, obj2_points[7]);
			break;
		case square :
			Square_Combine(obj->object2, obj2_points);
			break;
		case sphere:
		case cube:
		case cylinder:
			Cube_Combine(obj->object2, obj2_points);
			break;
		case cone:
			Pyramid_Combine(obj->object2, obj2_points);
			break;
		case torus:
			Torus_Combine(obj->object2, obj2_points);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			Compound_Combine(obj->object2, obj2_points);
			break;
	}

	/* Get the extents for both the intersecting bodies. */
	min_space = max_space = obj1_points[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (obj1_points[i].x > max_space.x)
			max_space.x = obj1_points[i].x;
		else if (obj1_points[i].x < min_space.x)
			min_space.x = obj1_points[i].x;
		if (obj1_points[i].y > max_space.y)
			max_space.y = obj1_points[i].y;
		else if (obj1_points[i].y < min_space.y)
			min_space.y = obj1_points[i].y;
		if (obj1_points[i].z > max_space.z)
			max_space.z = obj1_points[i].z;
		else if (obj1_points[i].z < min_space.z)
			min_space.z = obj1_points[i].z;
	}
	obj->object1->space_extent.max = max_space;
	obj->object1->space_extent.min = min_space;
	
	min_space = max_space = obj2_points[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (obj2_points[i].x > max_space.x)
			max_space.x = obj2_points[i].x;
		else if (obj2_points[i].x < min_space.x)
			min_space.x = obj2_points[i].x;
		if (obj2_points[i].y > max_space.y)
			max_space.y = obj2_points[i].y;
		else if (obj2_points[i].y < min_space.y)
			min_space.y = obj2_points[i].y;
		if (obj2_points[i].z > max_space.z)
			max_space.z = obj2_points[i].z;
		else if (obj2_points[i].z < min_space.z)
			min_space.z = obj2_points[i].z;
	}
	obj->object2->space_extent.max = max_space;
	obj->object2->space_extent.min = min_space;

	/* Transform all the points from the children. */
	/* This is computationally EXPENSIVE. */
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(obj->transform.matrix, obj1_points[i], temp_v);
		VAdd(obj->transform.displacement, temp_v, obj1_points[i]);
		MVMul(obj->transform.matrix, obj2_points[i], temp_v);
		VAdd(obj->transform.displacement, temp_v, obj2_points[i]);
	}


	if (obj->type == difference_op)
	{
		min_space = max_space = obj1_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj1_points[i].x > max_space.x)
				max_space.x = obj1_points[i].x;
			else if (obj1_points[i].x < min_space.x)
				min_space.x = obj1_points[i].x;
			if (obj1_points[i].y > max_space.y)
				max_space.y = obj1_points[i].y;
			else if (obj1_points[i].y < min_space.y)
				min_space.y = obj1_points[i].y;
			if (obj1_points[i].z > max_space.z)
				max_space.z = obj1_points[i].z;
			else if (obj1_points[i].z < min_space.z)
				min_space.z = obj1_points[i].z;
		}
		obj->space_extent.max = max_space;
		obj->space_extent.min = min_space;
	
		/* Put these same corners into view. */
		/* Convert all the corners to perspective view coords.	*/
		for ( i = 0 ; i < 8 ; i++ )
		{
			MVMul(w_to_v, obj1_points[i], view_points[i]);
			VAdd(disp, view_points[i], view_points[i]);
			temp_d = eyedist / (eyedist + view_points[i].z);
			view_points[i].x *= temp_d;
			view_points[i].y *= temp_d;
		}

		/* Get the bounding box. */
		obj->proj_extent.min.x = view_points[0].x;
		obj->proj_extent.min.y = view_points[0].y;
		obj->proj_extent.max.x = view_points[0].x;
		obj->proj_extent.max.y = view_points[0].y;
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj->proj_extent.min.x > view_points[i].x)
				obj->proj_extent.min.x = view_points[i].x;
			else if (obj->proj_extent.max.x < view_points[i].x)
				obj->proj_extent.max.x = view_points[i].x;
			if (obj->proj_extent.min.y > view_points[i].y)
				obj->proj_extent.min.y = view_points[i].y;
			else if (obj->proj_extent.max.y < view_points[i].y)
				obj->proj_extent.max.y = view_points[i].y;
		}

		return;
	}

	if (obj->type == union_op)
	{
		min_space = max_space = obj1_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj1_points[i].x > max_space.x)
				max_space.x = obj1_points[i].x;
			else if (obj1_points[i].x < min_space.x)
				min_space.x = obj1_points[i].x;
			if (obj1_points[i].y > max_space.y)
				max_space.y = obj1_points[i].y;
			else if (obj1_points[i].y < min_space.y)
				min_space.y = obj1_points[i].y;
			if (obj1_points[i].z > max_space.z)
				max_space.z = obj1_points[i].z;
			else if (obj1_points[i].z < min_space.z)
				min_space.z = obj1_points[i].z;
		}
		for ( i = 0 ; i < 8 ; i++ )
		{
			if (obj2_points[i].x > max_space.x)
				max_space.x = obj2_points[i].x;
			else if (obj2_points[i].x < min_space.x)
				min_space.x = obj2_points[i].x;
			if (obj2_points[i].y > max_space.y)
				max_space.y = obj2_points[i].y;
			else if (obj2_points[i].y < min_space.y)
				min_space.y = obj2_points[i].y;
			if (obj2_points[i].z > max_space.z)
				max_space.z = obj2_points[i].z;
			else if (obj2_points[i].z < min_space.z)
				min_space.z = obj2_points[i].z;
		}
		obj->space_extent.max = max_space;
		obj->space_extent.min = min_space;

		/* Put these same corners into view. */
		/* Convert all the corners to perspective view coords.	*/
		for ( i = 0 ; i < 8 ; i++ )
		{
			MVMul(w_to_v, obj1_points[i], view_points[i]);
			VAdd(disp, view_points[i], view_points[i]);
			temp_d = eyedist / (eyedist + view_points[i].z);
			view_points[i].x *= temp_d;
			view_points[i].y *= temp_d;
			MVMul(w_to_v, obj2_points[i], view_points[i+8]);
			VAdd(disp, view_points[i+8], view_points[i+8]);
			temp_d = eyedist / (eyedist + view_points[i+8].z);
			view_points[i+8].x *= temp_d;
			view_points[i+8].y *= temp_d;
		}

		/* Get the bounding box. */
		obj->proj_extent.min.x = view_points[0].x;
		obj->proj_extent.min.y = view_points[0].y;
		obj->proj_extent.max.x = view_points[0].x;
		obj->proj_extent.max.y = view_points[0].y;
		for ( i = 1 ; i < 16 ; i++ )
		{
			if (obj->proj_extent.min.x > view_points[i].x)
				obj->proj_extent.min.x = view_points[i].x;
			else if (obj->proj_extent.max.x < view_points[i].x)
				obj->proj_extent.max.x = view_points[i].x;
			if (obj->proj_extent.min.y > view_points[i].y)
				obj->proj_extent.min.y = view_points[i].y;
			else if (obj->proj_extent.max.y < view_points[i].y)
				obj->proj_extent.max.y = view_points[i].y;
		}

		return;
	}

	if (obj->type == intersection_op)
	{
		/* Get the extents for both the intersecting bodies. */
		min_space = max_space = obj1_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj1_points[i].x > max_space.x)
				max_space.x = obj1_points[i].x;
			else if (obj1_points[i].x < min_space.x)
				min_space.x = obj1_points[i].x;
			if (obj1_points[i].y > max_space.y)
				max_space.y = obj1_points[i].y;
			else if (obj1_points[i].y < min_space.y)
				min_space.y = obj1_points[i].y;
			if (obj1_points[i].z > max_space.z)
				max_space.z = obj1_points[i].z;
			else if (obj1_points[i].z < min_space.z)
				min_space.z = obj1_points[i].z;
		}
		obj->space_extent.max = max_space;
		obj->space_extent.min = min_space;
	
		min_space = max_space = obj2_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj2_points[i].x > max_space.x)
				max_space.x = obj2_points[i].x;
			else if (obj2_points[i].x < min_space.x)
				min_space.x = obj2_points[i].x;
			if (obj2_points[i].y > max_space.y)
				max_space.y = obj2_points[i].y;
			else if (obj2_points[i].y < min_space.y)
				min_space.y = obj2_points[i].y;
			if (obj2_points[i].z > max_space.z)
				max_space.z = obj2_points[i].z;
			else if (obj2_points[i].z < min_space.z)
				min_space.z = obj2_points[i].z;
		}

		obj->space_extent.max.x = min(obj->space_extent.max.x, max_space.x);
		obj->space_extent.max.y = min(obj->space_extent.max.y, max_space.y);
		obj->space_extent.max.z = min(obj->space_extent.max.z, max_space.z);
		obj->space_extent.min.x = max(obj->space_extent.min.x, min_space.x);
		obj->space_extent.min.y = max(obj->space_extent.min.y, min_space.y);
		obj->space_extent.min.z = max(obj->space_extent.min.z, min_space.z);

		/* Convert all the corners to perspective view coords.	*/
		for ( i = 0 ; i < 8 ; i++ )
		{
			MVMul(w_to_v, obj1_points[i], view_points[i]);
			VAdd(disp, view_points[i], view_points[i]);
			temp_d = eyedist / (eyedist + view_points[i].z);
			view_points[i].x *= temp_d;
			view_points[i].y *= temp_d;
			MVMul(w_to_v, obj2_points[i], view_points[i+8]);
			VAdd(disp, view_points[i+8], view_points[i+8]);
			temp_d = eyedist / (eyedist + view_points[i+8].z);
			view_points[i+8].x *= temp_d;
			view_points[i+8].y *= temp_d;
		}

		/* Get the bounding box for each object. */
		obj->object1->proj_extent.min.x = view_points[0].x;
		obj->object1->proj_extent.min.y = view_points[0].y;
		obj->object1->proj_extent.max.x = view_points[0].x;
		obj->object1->proj_extent.max.y = view_points[0].y;
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj->object1->proj_extent.min.x > view_points[i].x)
				obj->object1->proj_extent.min.x = view_points[i].x;
			else if (obj->object1->proj_extent.max.x < view_points[i].x)
				obj->object1->proj_extent.max.x = view_points[i].x;
			if (obj->object1->proj_extent.min.y > view_points[i].y)
				obj->object1->proj_extent.min.y = view_points[i].y;
			else if (obj->object1->proj_extent.max.y < view_points[i].y)
				obj->object1->proj_extent.max.y = view_points[i].y;
		}
		obj->object2->proj_extent.min.x = view_points[8].x;
		obj->object2->proj_extent.min.y = view_points[8].y;
		obj->object2->proj_extent.max.x = view_points[8].x;
		obj->object2->proj_extent.max.y = view_points[8].y;
		for ( i = 9 ; i < 16 ; i++ )
		{
			if (obj->object2->proj_extent.min.x > view_points[i].x)
				obj->object2->proj_extent.min.x = view_points[i].x;
			else if (obj->object2->proj_extent.max.x < view_points[i].x)
				obj->object2->proj_extent.max.x = view_points[i].x;
			if (obj->object2->proj_extent.min.y > view_points[i].y)
				obj->object2->proj_extent.min.y = view_points[i].y;
			else if (obj->object2->proj_extent.max.y < view_points[i].y)
				obj->object2->proj_extent.max.y = view_points[i].y;
		}

		/* Now process the intersection. */
		obj->proj_extent.max.x = min(obj->object1->proj_extent.max.x,
									 obj->object2->proj_extent.max.x);
		obj->proj_extent.max.y = min(obj->object1->proj_extent.max.y,
									 obj->object2->proj_extent.max.y);
		obj->proj_extent.min.x = max(obj->object1->proj_extent.min.x,
									 obj->object2->proj_extent.min.x);
		obj->proj_extent.min.y = max(obj->object1->proj_extent.min.y,
									 obj->object2->proj_extent.min.y);

		return;
	}
}


void
Square_Extent(Instance *obj, Vector* w_to_v, Vector disp)
/************************************************************************/
/*																		*/
/*	Determines the extents for a square.								*/
/*	Basically takes the corners of a generic square, transforms			*/
/*	 them to view and gets the x & y projections onto the screen.		*/
/*																		*/
/************************************************************************/
{
	Vector	square_world[4];
	Vector	square_view[4];
	double	temp_d;
	Vector	min_space, max_space;

	/* Get the corners in world.	*/
	/* This is ugly but works because the corners all have 1 in there somewhere.	*/
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y +
		 obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y +
		 obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y +
		 obj->transform.displacement.z,
	     square_world[0]);
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y +
		 obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y +
		 obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y +
		 obj->transform.displacement.z,
	     square_world[1]);
	VNew(-(obj->transform.matrix[0].x) - obj->transform.matrix[0].y +
		  obj->transform.displacement.x,
	     -(obj->transform.matrix[1].x) - obj->transform.matrix[1].y +
		  obj->transform.displacement.y,
	     -(obj->transform.matrix[2].x) - obj->transform.matrix[2].y +
		  obj->transform.displacement.z,
	     square_world[2]);
	VNew(-(obj->transform.matrix[0].x) + obj->transform.matrix[0].y +
		  obj->transform.displacement.x,
	     -(obj->transform.matrix[1].x) + obj->transform.matrix[1].y +
		  obj->transform.displacement.y,
	     -(obj->transform.matrix[2].x) + obj->transform.matrix[2].y +
		  obj->transform.displacement.z,
	     square_world[3]);
	
	/* Save the spacial extents for secondary ray casting.	*/
	min_space = max_space = square_world[0];
	for ( i = 1 ; i < 4 ; i++ )
	{
		if (square_world[i].x > max_space.x) max_space.x = square_world[i].x;
		else if (square_world[i].x < min_space.x) min_space.x = square_world[i].x;
		if (square_world[i].y > max_space.y) max_space.y = square_world[i].y;
		else if (square_world[i].y < min_space.y) min_space.y = square_world[i].y;
		if (square_world[i].z > max_space.z) max_space.z = square_world[i].z;
		else if (square_world[i].z < min_space.z) min_space.z = square_world[i].z;
	}
	obj->space_extent.max = max_space;
	obj->space_extent.min = min_space;
	
	/* Put these same corners into view. */
	MVMul(w_to_v, square_world[0], square_view[0]);
	MVMul(w_to_v, square_world[1], square_view[1]);
	MVMul(w_to_v, square_world[2], square_view[2]);
	MVMul(w_to_v, square_world[3], square_view[3]);

	VAdd(disp, square_view[0], square_view[0]);
	VAdd(disp, square_view[1], square_view[1]);
	VAdd(disp, square_view[2], square_view[2]);
	VAdd(disp, square_view[3], square_view[3]);

	temp_d = eyedist / (eyedist + square_view[0].z);
	square_view[0].x *= temp_d;
	square_view[0].y *= temp_d;
	temp_d = eyedist / (eyedist + square_view[1].z);
	square_view[1].x *= temp_d;
	square_view[1].y *= temp_d;
	temp_d = eyedist / (eyedist + square_view[2].z);
	square_view[2].x *= temp_d;
	square_view[2].y *= temp_d;
	temp_d = eyedist / (eyedist + square_view[3].z);
	square_view[3].x *= temp_d;
	square_view[3].y *= temp_d;

	/* Get the bounding box. */
	obj->proj_extent.min.x = square_view[0].x;
	obj->proj_extent.min.y = square_view[0].y;
	obj->proj_extent.max.x = square_view[0].x;
	obj->proj_extent.max.y = square_view[0].y;
	for ( i = 1 ; i < 4 ; i++ )
	{
		if (obj->proj_extent.min.x > square_view[i].x)
			obj->proj_extent.min.x = square_view[i].x;
		else if (obj->proj_extent.max.x < square_view[i].x)
			obj->proj_extent.max.x = square_view[i].x;
		if (obj->proj_extent.min.y > square_view[i].y)
			obj->proj_extent.min.y = square_view[i].y;
		else if (obj->proj_extent.max.y < square_view[i].y)
			obj->proj_extent.max.y = square_view[i].y;
	}
}


void
Cube_Extent(Instance *obj, Vector* w_to_v, Vector disp)
/************************************************************************/
/*																		*/
/*	Determines the extents for a cube.									*/
/*	Method as for the square.											*/
/*																		*/
/************************************************************************/
{
	Vector	cube_world[8];
	Vector	cube_view[8];
	Vector	min_space, max_space;
	double	temp_d;


	/* Get the corners in world.	*/
	/* This is ugly but works because the corners all have 1 in there somewhere.	*/
	/* (1,1,1) */
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[0]);

	/* (1,1,-1) */
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[1]);

	/* (1,-1,1) */
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[2]);

	/* (1,-1,-1) */
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[3]);

	/* (-1,1,1) */
	VNew(- obj->transform.matrix[0].x + obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x + obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x + obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[4]);

	/* (-1,1,-1) */
	VNew(- obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[5]);

	/* (-1,-1,1) */
	VNew(- obj->transform.matrix[0].x - obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x - obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x - obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[6]);

	/* (-1,-1,-1) */
	VNew(- obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     cube_world[7]);

	
	/* Save the spacial extents for secondary ray casting.	*/
	min_space = max_space = cube_world[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (cube_world[i].x > max_space.x) max_space.x = cube_world[i].x;
		else if (cube_world[i].x < min_space.x) min_space.x = cube_world[i].x;
		if (cube_world[i].y > max_space.y) max_space.y = cube_world[i].y;
		else if (cube_world[i].y < min_space.y) min_space.y = cube_world[i].y;
		if (cube_world[i].z > max_space.z) max_space.z = cube_world[i].z;
		else if (cube_world[i].z < min_space.z) min_space.z = cube_world[i].z;
	}
	obj->space_extent.max = max_space;
	obj->space_extent.min = min_space;
	

	/* Convert all the corners to perspective view coords.	*/
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(w_to_v, cube_world[i], cube_view[i]);
		VAdd(disp, cube_view[i], cube_view[i]);
		temp_d = eyedist / (eyedist + cube_view[i].z);
		cube_view[i].x *= temp_d;
		cube_view[i].y *= temp_d;
	}

	/* Get the bounding box. */
	obj->proj_extent.min.x = cube_view[0].x;
	obj->proj_extent.min.y = cube_view[0].y;
	obj->proj_extent.max.x = cube_view[0].x;
	obj->proj_extent.max.y = cube_view[0].y;
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (obj->proj_extent.min.x > cube_view[i].x)
			obj->proj_extent.min.x = cube_view[i].x;
		else if (obj->proj_extent.max.x < cube_view[i].x)
			obj->proj_extent.max.x = cube_view[i].x;
		if (obj->proj_extent.min.y > cube_view[i].y)
			obj->proj_extent.min.y = cube_view[i].y;
		else if (obj->proj_extent.max.y < cube_view[i].y)
			obj->proj_extent.max.y = cube_view[i].y;
	}
}


void
Pyramid_Extent(Instance *obj, Vector* w_to_v, Vector disp)
/************************************************************************/
/*																		*/
/*	Determines the extents for a tetrahedron.							*/
/*	Method as for the square.											*/
/*																		*/
/************************************************************************/
{
	Vector	pyr_world[5];
	Vector	pyr_view[5];
	Vector	min_space, max_space;
	double	temp_d;

	/* Get the corners in world.	*/
	/* This is ugly but works because the corners all have 1 in there somewhere.	*/
	/* (0,0,1) */
	VNew(obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].z + obj->transform.displacement.z,
	     pyr_world[0]);

	/* (1,1,-1) */
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     pyr_world[1]);

	/* (1,-1,-1) */
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     pyr_world[2]);

	/* (-1,1,-1) */
	VNew(- obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     pyr_world[3]);

	/* (-1,-1,-1) */
	VNew(- obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     pyr_world[4]);

	
	/* Save the spacial extents for secondary ray casting.	*/
	min_space = max_space = pyr_world[0];
	for ( i = 1 ; i < 5 ; i++ )
	{
		if (pyr_world[i].x > max_space.x) max_space.x = pyr_world[i].x;
		else if (pyr_world[i].x < min_space.x) min_space.x = pyr_world[i].x;
		if (pyr_world[i].y > max_space.y) max_space.y = pyr_world[i].y;
		else if (pyr_world[i].y < min_space.y) min_space.y = pyr_world[i].y;
		if (pyr_world[i].z > max_space.z) max_space.z = pyr_world[i].z;
		else if (pyr_world[i].z < min_space.z) min_space.z = pyr_world[i].z;
	}
	obj->space_extent.max = max_space;
	obj->space_extent.min = min_space;
	

	/* Convert all the corners to perspective view coords.	*/
	for ( i = 0 ; i < 5 ; i++ )
	{
		MVMul(w_to_v, pyr_world[i], pyr_view[i]);
		VAdd(disp, pyr_view[i], pyr_view[i]);
		temp_d = eyedist / (eyedist + pyr_view[i].z);
		pyr_view[i].x *= temp_d;
		pyr_view[i].y *= temp_d;
	}

	/* Get the bounding box. */
	obj->proj_extent.min.x = pyr_view[0].x;
	obj->proj_extent.min.y = pyr_view[0].y;
	obj->proj_extent.max.x = pyr_view[0].x;
	obj->proj_extent.max.y = pyr_view[0].y;
	for ( i = 1 ; i < 5 ; i++ )
	{
		if (obj->proj_extent.min.x > pyr_view[i].x)
			obj->proj_extent.min.x = pyr_view[i].x;
		else if (obj->proj_extent.max.x < pyr_view[i].x)
			obj->proj_extent.max.x = pyr_view[i].x;
		if (obj->proj_extent.min.y > pyr_view[i].y)
			obj->proj_extent.min.y = pyr_view[i].y;
		else if (obj->proj_extent.max.y < pyr_view[i].y)
			obj->proj_extent.max.y = pyr_view[i].y;
	}
}


void
Torus_Extent(Instance *obj, Vector* w_to_v, Vector disp)
/************************************************************************/
/*																		*/
/*	Determines the extents for a torus.									*/
/*	First generates the bounding box in generic, then in world, then	*/
/*	 in view.															*/
/*																		*/
/************************************************************************/
{
	Vector	torus_generic[8];
	Vector	torus_world[8];
	Vector	torus_view[8];
	Vector	min_space, max_space;
	double	temp_d;


	/* Get the corners in generic.	*/
	temp_d = obj->torus_rad + 1.0;
	VNew(temp_d, temp_d, 1, torus_generic[0]);
	VNew(temp_d, temp_d, -1, torus_generic[1]);
	VNew(temp_d, -temp_d, 1, torus_generic[2]);
	VNew(temp_d, -temp_d, -1, torus_generic[3]);
	VNew(-temp_d, temp_d, 1, torus_generic[4]);
	VNew(-temp_d, temp_d, -1, torus_generic[5]);
	VNew(-temp_d, -temp_d, 1, torus_generic[6]);
	VNew(-temp_d, -temp_d, -1, torus_generic[7]);


	/* Transform all the corners into world. */
	/* Transform all the points from the children. */
	/* This is computationally EXPENSIVE. */
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(obj->transform.matrix, torus_generic[i], torus_world[i]);
		VAdd(obj->transform.displacement, torus_world[i], torus_world[i]);
	}

	/* Take the max and min across all the points. */
	min_space = max_space = torus_world[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (torus_world[i].x > max_space.x)
			max_space.x = torus_world[i].x;
		else if (torus_world[i].x < min_space.x)
			min_space.x = torus_world[i].x;
		if (torus_world[i].y > max_space.y)
			max_space.y = torus_world[i].y;
		else if (torus_world[i].y < min_space.y)
			min_space.y = torus_world[i].y;
		if (torus_world[i].z > max_space.z)
			max_space.z = torus_world[i].z;
		else if (torus_world[i].z < min_space.z)
			min_space.z = torus_world[i].z;
	}
	obj->space_extent.max = max_space;
	obj->space_extent.min = min_space;

	/* Convert all the corners to perspective view coords.	*/
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(w_to_v, torus_world[i], torus_view[i]);
		VAdd(disp, torus_view[i], torus_view[i]);
		temp_d = eyedist / (eyedist + torus_view[i].z);
		torus_view[i].x *= temp_d;
		torus_view[i].y *= temp_d;
	}

	/* Get the bounding box. */
	obj->proj_extent.min.x = torus_view[0].x;
	obj->proj_extent.min.y = torus_view[0].y;
	obj->proj_extent.max.x = torus_view[0].x;
	obj->proj_extent.max.y = torus_view[0].y;
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (obj->proj_extent.min.x > torus_view[i].x)
			obj->proj_extent.min.x = torus_view[i].x;
		else if (obj->proj_extent.max.x < torus_view[i].x)
			obj->proj_extent.max.x = torus_view[i].x;
		if (obj->proj_extent.min.y > torus_view[i].y)
			obj->proj_extent.min.y = torus_view[i].y;
		else if (obj->proj_extent.max.y < torus_view[i].y)
			obj->proj_extent.max.y = torus_view[i].y;
	}
}


void
Compound_Combine(Instance *obj, Vector *points)
/****************************************************************/
/*																*/
/*	The sub-extent function for compound objects.  Like			*/
/*	 Compound_Extent but without the proj_extent calculations.	*/
/*																*/
/****************************************************************/
{
	Vector	obj1_points[8];
	Vector	obj2_points[8];
	Vector	min_space, max_space;
	Vector	temp_v;

	/* Get the corners for each child. */
	switch (obj->object1->type)
	{
		case plane :
			/* The person specifying such an object is crazy. */
			/* The projection and spacial extent for the	*/
			/*	 whole compound object will be the whole	*/
			/*	 of space.									*/
			VNew(HUGE, HUGE, HUGE, obj1_points[0]);
			VNew(HUGE, HUGE, -HUGE, obj1_points[1]);
			VNew(HUGE, -HUGE, HUGE, obj1_points[2]);
			VNew(HUGE, -HUGE, -HUGE, obj1_points[3]);
			VNew(-HUGE, HUGE, HUGE, obj1_points[4]);
			VNew(-HUGE, HUGE, -HUGE, obj1_points[5]);
			VNew(-HUGE, -HUGE, HUGE, obj1_points[6]);
			VNew(-HUGE, -HUGE, -HUGE, obj1_points[7]);
			break;
		case square :
			Square_Combine(obj->object1, obj1_points);
			break;
		case sphere:
		case cube:
		case cylinder:
			Cube_Combine(obj->object1, obj1_points);
			break;
		case cone:
			Pyramid_Combine(obj->object1, obj1_points);
			break;
		case torus:
			Torus_Combine(obj->object1, obj1_points);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			Compound_Combine(obj->object1, obj1_points);
			break;
	}

	switch (obj->object2->type)
	{
		case plane :
			/* The person specifying such an object is crazy. */
			/* The projection and spacial extent for the	*/
			/*	 whole compound object will be the whole	*/
			/*	 of space.									*/
			VNew(HUGE, HUGE, HUGE, obj2_points[0]);
			VNew(HUGE, HUGE, -HUGE, obj2_points[1]);
			VNew(HUGE, -HUGE, HUGE, obj2_points[2]);
			VNew(HUGE, -HUGE, -HUGE, obj2_points[3]);
			VNew(-HUGE, HUGE, HUGE, obj2_points[4]);
			VNew(-HUGE, HUGE, -HUGE, obj2_points[5]);
			VNew(-HUGE, -HUGE, HUGE, obj2_points[6]);
			VNew(-HUGE, -HUGE, -HUGE, obj2_points[7]);
			break;
		case square :
			Square_Combine(obj->object2, obj2_points);
			break;
		case sphere:
		case cube:
		case cylinder:
			Cube_Combine(obj->object2, obj2_points);
			break;
		case cone:
			Pyramid_Combine(obj->object2, obj2_points);
			break;
		case torus:
			Torus_Combine(obj->object2, obj2_points);
			break;
		case union_op:
		case intersection_op:
		case difference_op:
			Compound_Combine(obj->object2, obj2_points);
			break;
	}

	/* Get the extents for both the intersecting bodies. */
	min_space = max_space = obj1_points[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (obj1_points[i].x > max_space.x)
			max_space.x = obj1_points[i].x;
		else if (obj1_points[i].x < min_space.x)
			min_space.x = obj1_points[i].x;
		if (obj1_points[i].y > max_space.y)
			max_space.y = obj1_points[i].y;
		else if (obj1_points[i].y < min_space.y)
			min_space.y = obj1_points[i].y;
		if (obj1_points[i].z > max_space.z)
			max_space.z = obj1_points[i].z;
		else if (obj1_points[i].z < min_space.z)
			min_space.z = obj1_points[i].z;
	}
	obj->object1->space_extent.max = max_space;
	obj->object1->space_extent.min = min_space;
	
	min_space = max_space = obj2_points[0];
	for ( i = 1 ; i < 8 ; i++ )
	{
		if (obj2_points[i].x > max_space.x)
			max_space.x = obj2_points[i].x;
		else if (obj2_points[i].x < min_space.x)
			min_space.x = obj2_points[i].x;
		if (obj2_points[i].y > max_space.y)
			max_space.y = obj2_points[i].y;
		else if (obj2_points[i].y < min_space.y)
			min_space.y = obj2_points[i].y;
		if (obj2_points[i].z > max_space.z)
			max_space.z = obj2_points[i].z;
		else if (obj2_points[i].z < min_space.z)
			min_space.z = obj2_points[i].z;
	}
	obj->object2->space_extent.max = max_space;
	obj->object2->space_extent.min = min_space;

	/* Transform all the points from the children. */
	/* This is computationally EXPENSIVE. */
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(obj->transform.matrix, obj1_points[i], temp_v);
		VAdd(obj->transform.displacement, temp_v, temp_v);
		obj1_points[i] = temp_v;
		MVMul(obj->transform.matrix, obj2_points[i], temp_v);
		VAdd(obj->transform.displacement, temp_v, temp_v);
		obj2_points[i] = temp_v;
	}


	if (obj->type == difference_op)
	{
		for ( i = 0 ; i < 8 ; i++ )
			points[i] = obj1_points[i];

		return;
	}

	if (obj->type == union_op)
	{
		min_space = max_space = obj1_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj1_points[i].x > max_space.x)
				max_space.x = obj1_points[i].x;
			else if (obj1_points[i].x < min_space.x)
				min_space.x = obj1_points[i].x;
			if (obj1_points[i].y > max_space.y)
				max_space.y = obj1_points[i].y;
			else if (obj1_points[i].y < min_space.y)
				min_space.y = obj1_points[i].y;
			if (obj1_points[i].z > max_space.z)
				max_space.z = obj1_points[i].z;
			else if (obj1_points[i].z < min_space.z)
				min_space.z = obj1_points[i].z;
		}
		for ( i = 0 ; i < 8 ; i++ )
		{
			if (obj2_points[i].x > max_space.x)
				max_space.x = obj2_points[i].x;
			else if (obj2_points[i].x < min_space.x)
				min_space.x = obj2_points[i].x;
			if (obj2_points[i].y > max_space.y)
				max_space.y = obj2_points[i].y;
			else if (obj2_points[i].y < min_space.y)
				min_space.y = obj2_points[i].y;
			if (obj2_points[i].z > max_space.z)
				max_space.z = obj2_points[i].z;
			else if (obj2_points[i].z < min_space.z)
				min_space.z = obj2_points[i].z;
		}

		points[0] = max_space;
		VNew(max_space.x, max_space.y, min_space.z, points[1]);
		VNew(max_space.x, min_space.y, max_space.z, points[2]);
		VNew(max_space.x, min_space.y, min_space.z, points[3]);
		VNew(min_space.x, max_space.y, max_space.z, points[4]);
		VNew(min_space.x, max_space.y, min_space.z, points[5]);
		VNew(min_space.x, min_space.y, max_space.z, points[6]);
		points[7] = min_space;
	
		return;
	}

	if (obj->type == intersection_op)
	{
		/* Get the extents for both the intersecting bodies. */
		min_space = max_space = obj1_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj1_points[i].x > max_space.x)
				max_space.x = obj1_points[i].x;
			else if (obj1_points[i].x < min_space.x)
				min_space.x = obj1_points[i].x;
			if (obj1_points[i].y > max_space.y)
				max_space.y = obj1_points[i].y;
			else if (obj1_points[i].y < min_space.y)
				min_space.y = obj1_points[i].y;
			if (obj1_points[i].z > max_space.z)
				max_space.z = obj1_points[i].z;
			else if (obj1_points[i].z < min_space.z)
				min_space.z = obj1_points[i].z;
		}
		obj->space_extent.max = max_space;
		obj->space_extent.min = min_space;
	
		min_space = max_space = obj2_points[0];
		for ( i = 1 ; i < 8 ; i++ )
		{
			if (obj2_points[i].x > max_space.x)
				max_space.x = obj2_points[i].x;
			else if (obj2_points[i].x < min_space.x)
				min_space.x = obj2_points[i].x;
			if (obj2_points[i].y > max_space.y)
				max_space.y = obj2_points[i].y;
			else if (obj2_points[i].y < min_space.y)
				min_space.y = obj2_points[i].y;
			if (obj2_points[i].z > max_space.z)
				max_space.z = obj2_points[i].z;
			else if (obj2_points[i].z < min_space.z)
				min_space.z = obj2_points[i].z;
		}

		max_space.x = min(obj->space_extent.max.x, max_space.x);
		max_space.y = min(obj->space_extent.max.y, max_space.y);
		max_space.z = min(obj->space_extent.max.z, max_space.z);
		min_space.x = max(obj->space_extent.min.x, min_space.x);
		min_space.y = max(obj->space_extent.min.y, min_space.y);
		min_space.z = max(obj->space_extent.min.z, min_space.z);

		points[0] = max_space;
		VNew(max_space.x, max_space.y, min_space.z, points[1]);
		VNew(max_space.x, min_space.y, max_space.z, points[2]);
		VNew(max_space.x, min_space.y, min_space.z, points[3]);
		VNew(min_space.x, max_space.y, max_space.z, points[4]);
		VNew(min_space.x, max_space.y, min_space.z, points[5]);
		VNew(min_space.x, min_space.y, max_space.z, points[6]);
		points[7] = min_space;
	
		return;
	}
}



void
Square_Combine(Instance *obj, Vector *points)
/****************************************************************/
/*																*/
/*	The function which calculates the 8 bounding points for a	*/
/*	 square.  A simplified version of sqaure_extent, basically.	*/
/*																*/
/****************************************************************/
{

	/* Get the corners in world.	*/
	/* This is ugly but works because the corners all have 1 in there somewhere.	*/
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y +
		 obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y +
		 obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y +
		 obj->transform.displacement.z,
	     points[0]);
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y +
		 obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y +
		 obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y +
		 obj->transform.displacement.z,
	     points[2]);
	VNew(-(obj->transform.matrix[0].x) - obj->transform.matrix[0].y +
		  obj->transform.displacement.x,
	     -(obj->transform.matrix[1].x) - obj->transform.matrix[1].y +
		  obj->transform.displacement.y,
	     -(obj->transform.matrix[2].x) - obj->transform.matrix[2].y +
		  obj->transform.displacement.z,
	     points[6]);
	VNew(-(obj->transform.matrix[0].x) + obj->transform.matrix[0].y +
		  obj->transform.displacement.x,
	     -(obj->transform.matrix[1].x) + obj->transform.matrix[1].y +
		  obj->transform.displacement.y,
	     -(obj->transform.matrix[2].x) + obj->transform.matrix[2].y +
		  obj->transform.displacement.z,
	     points[4]);

	points[1] = points[0];
	points[3] = points[2];
	points[7] = points[6];
	points[5] = points[4];

}



void
Cube_Combine(Instance *obj, Vector *points)
/****************************************************************/
/*																*/
/*	The function which calculates the 8 bounding points for a	*/
/*	 cube.  A simplified version of Cube_Extent, basically.		*/
/*																*/
/****************************************************************/
{

	/* Get the corners in world.	*/
	/* (1,1,1) */
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[0]);

	/* (1,1,-1) */
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[1]);

	/* (1,-1,1) */
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[2]);

	/* (1,-1,-1) */
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[3]);

	/* (-1,1,1) */
	VNew(- obj->transform.matrix[0].x + obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x + obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x + obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[4]);

	/* (-1,1,-1) */
	VNew(- obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[5]);

	/* (-1,-1,1) */
	VNew(- obj->transform.matrix[0].x - obj->transform.matrix[0].y +
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x - obj->transform.matrix[1].y +
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x - obj->transform.matrix[2].y +
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[6]);

	/* (-1,-1,-1) */
	VNew(- obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[7]);

}



void
Pyramid_Combine(Instance *obj, Vector *points)
/****************************************************************/
/*																*/
/*	The function which calculates the 8 bounding points for a	*/
/*	 pyramid.  A simplified version of Pyramid_Extent,			*/
/*	 basically.													*/
/*																*/
/****************************************************************/
{
	/* Get the corners in world.	*/
	/* (0,0,1) */
	VNew(obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[0]);
	points[2] = points[6] = points[4] = points[0];

	/* (1,1,-1) */
	VNew(obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[1]);

	/* (1,-1,-1) */
	VNew(obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[3]);

	/* (-1,1,-1) */
	VNew(- obj->transform.matrix[0].x + obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x + obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x + obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[5]);

	/* (-1,-1,-1) */
	VNew(- obj->transform.matrix[0].x - obj->transform.matrix[0].y -
		obj->transform.matrix[0].z + obj->transform.displacement.x,
	     - obj->transform.matrix[1].x - obj->transform.matrix[1].y -
		obj->transform.matrix[1].z + obj->transform.displacement.y,
	     - obj->transform.matrix[2].x - obj->transform.matrix[2].y -
		obj->transform.matrix[2].z + obj->transform.displacement.z,
	     points[7]);

	
}



void
Torus_Combine(Instance *obj, Vector *points)
/************************************************************************/
/*																		*/
/*	Determines the extents for a torus.									*/
/*	First generates the bounding box in generic, then in world, then	*/
/*	 in view.															*/
/*																		*/
/************************************************************************/
{
	Vector	torus_generic[8];
	double	temp_d;


	/* Get the corners in generic.	*/
	temp_d = obj->torus_rad + 1.0;
	VNew(temp_d, temp_d, 1, torus_generic[0]);
	VNew(temp_d, temp_d, -1, torus_generic[1]);
	VNew(temp_d, -temp_d, 1, torus_generic[2]);
	VNew(temp_d, -temp_d, -1, torus_generic[3]);
	VNew(-temp_d, temp_d, 1, torus_generic[4]);
	VNew(-temp_d, temp_d, -1, torus_generic[5]);
	VNew(-temp_d, -temp_d, 1, torus_generic[6]);
	VNew(-temp_d, -temp_d, -1, torus_generic[7]);


	/* Transform all the corners into world. */
	/* Transform all the points from the children. */
	/* This is computationally EXPENSIVE. */
	for ( i = 0 ; i < 8 ; i++ )
	{
		MVMul(obj->transform.matrix, torus_generic[i], points[i]);
		VAdd(obj->transform.displacement, points[i], points[i]);
	}
}



int
Ray_In_Extent(Ray ray, Instance *obj)
{
	double	tin = -HUGE;
	double	tout = HUGE;
	double	thit;

	/* Since there are so many 0's in all this, I'm going	*/
	/*  to do it linearly.  This makes for poor readability	*/
	/*  and lots of code, but much greater speed.		*/

	/* x = a plane: n = {1,0,0}, D = obj->space_extent.max.x */
	if (ray.direction.x == 0.0)	/* Ray parallel.	*/
	{
		if ((obj->space_extent.max.x - ray.start_point.x < 0.0) ||
			(-obj->space_extent.min.x + ray.start_point.x < 0.0))
			return FALSE;
	}
	else
	{
		thit = (obj->space_extent.max.x - ray.start_point.x) / ray.direction.x;
		if (ray.direction.x > 0.0)
			tout = thit;
		else
			tin = thit;
	}
	/* Not possible to eliminate ray yet */

	/* y = 1 plane: n = {0,1,0}, D = obj->space_extent.max.y */
	if (ray.direction.y == 0.0)     /* Ray parallel.        */
	{
		if ((obj->space_extent.max.y - ray.start_point.y < 0.0) ||
			(-obj->space_extent.min.y + ray.start_point.y < 0.0))
			return FALSE;
	}
	else
	{
		thit = (obj->space_extent.max.y - ray.start_point.y) / ray.direction.y;
		if (ray.direction.y > 0.0)
		{
			if (thit < tout) tout = thit;
		}
		else
		{
			if (thit > tin) tin = thit;
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return FALSE;

	/* z = 1 plane: n = {0,0,1}, D = obj->space_extent.max.z */
	if (ray.direction.z == 0.0)     /* Ray parallel.        */
	{
		if ((obj->space_extent.max.z - ray.start_point.z < 0.0) ||
			(-obj->space_extent.min.z + ray.start_point.z < 0.0))
			return FALSE;
	}
	else
	{
		thit = (obj->space_extent.max.z - ray.start_point.z) / ray.direction.z;
		if (ray.direction.z > 0.0)
		{
			if (thit < tout) tout = thit;
		}
		else
		{
			if (thit > tin) tin = thit;
		}
	}
	/* Try to eliminate intersection. */
	if (tin > tout) return FALSE;


	/* x = -1 plane : n = {-1,0,0}, D = -obj->space_extent.min.x */
	/* Already eliminated parallel rays above. */
	thit = (-obj->space_extent.min.x + ray.start_point.x) / -ray.direction.x;
	if (ray.direction.x <  0.0)
	{
		if (thit < tout) tout = thit;
	}
	else
	{
		if (thit > tin) tin = thit;
	}
	if (tin > tout) return FALSE;

	/* y = -1 plane : n = {0,-1,0}, D = -obj->space_extent.min.y */
	/* Already eliminated parallel rays above. */
	thit = (-obj->space_extent.min.y + ray.start_point.y) / -ray.direction.y;
	if (ray.direction.y <  0.0)
	{
		if (thit < tout) tout = thit;
	}
	else
	{
		if (thit > tin) tin = thit;
	}
	if (tin > tout) return FALSE;

	/* z = -1 plane : n = {0,0,-1}, D = -obj->space_extent.min.z */
	/* Already eliminated parallel rays above. */
	thit = (-obj->space_extent.min.z + ray.start_point.z) / -ray.direction.z;
	if (ray.direction.z <  0.0)
	{
		if (thit < tout) tout = thit;
	}
	else
	{
		if (thit > tin) tin = thit;
	}
	if (tin > tout) return FALSE;

	/* Here if there is a valid intersection */
	if ((tin < DELTA) && (tout < DELTA)) return FALSE;

	return TRUE;

}

