/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/****************************************************************/
/*																*/
/*	memory.c : Specialised memory management functions for		*/
/*				genray.											*/
/*	Written by Stephen Chenney : Jan 1994.						*/
/*																*/
/****************************************************************/

#include "genray.h"


HitList	*first_hit_free = NULL;
HitList	*last_hit_free = NULL;
InstanceList	*first_inst_free = NULL;
InstanceList	*last_inst_free = NULL;

HitList*
get_hit_elmt()
/****************************************************************/
/*																*/
/*	Returns a new HitList object.								*/
/*																*/
/****************************************************************/
{
	HitList	*new;

	if (!first_hit_free)
	{

		/* None available.  Get a new one. */
		if ((new = (HitList*)malloc(sizeof(HitList))) == NULL)
		{
			fprintf(stderr, "genray: ERROR: Out of memory.\n");
			exit(1);
		}

		return new;
	}
	else
	{
		new = first_hit_free;
		first_hit_free = first_hit_free->next;
		return new;
	}
}

InstanceList*
get_inst_elmt()
/****************************************************************/
/*																*/
/*	Returns a new InstanceList object.							*/
/*																*/
/****************************************************************/
{
	InstanceList	*new;

	if (!first_inst_free)
	{

		/* None available.  Get a new one. */
		if ((new = (InstanceList*)malloc(sizeof(InstanceList))) == NULL)
		{
			fprintf(stderr, "genray: ERROR: Out of memory.\n");
			exit(1);
		}

		return new;
	}
	else
	{
		new = first_inst_free;
		first_inst_free = first_inst_free->next;
		return new;
	}
}


void
free_hit_elmt(HitList *elmt)
/****************************************************************/
/*																*/
/*	Frees a HitList object.									*/
/*																*/
/****************************************************************/
{
	if (!first_hit_free)
	{
		first_hit_free = last_hit_free = elmt;
	}
	else
	{
		last_hit_free->next = elmt;
		last_hit_free = elmt;
	}
	elmt->next = NULL;
}

void
free_inst_elmt(InstanceList *elmt)
/****************************************************************/
/*																*/
/*	Frees a InstanceList object.								*/
/*																*/
/****************************************************************/
{
	if (!first_inst_free)
	{
		first_inst_free = last_inst_free = elmt;
	}
	else
	{
		last_inst_free->next = elmt;
		last_inst_free = elmt;
	}
	elmt->next = NULL;
}

