/*
**    Genray: A Generic Raytracer
**    Copyright (C) 1994  Stephen Chenney (stephen@cs.su.oz.au)
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
**	rayin.l : lex input file for raytrace
**	Thanks to Franc Carter.
*/

%e 2000

%{

#include <stdlib.h>
#include <malloc.h>
#include "rayin.h"

#undef YY_DECL
#define YY_DECL input_token yylex()

void comment(void);
void read_string(void);

int    lex_int;
double lex_float;
char   *lex_string;

#ifdef FLEX
int		yylineno = 1;
#endif
char	*input_file_name;

/* The base filename for includes etc.  Defined in init.c */
extern	char	*file_dir;

%}


SPACE       ("\t"|" ")
A           ("a"|"A")
B           ("b"|"B")
C           ("c"|"C")
D           ("d"|"D")
E           ("e"|"E")
F           ("f"|"F")
G           ("g"|"G")
H           ("h"|"H")
I           ("i"|"I")
J           ("j"|"J")
K           ("k"|"K")
L           ("l"|"L")
M           ("m"|"M")
N           ("n"|"N")
O           ("o"|"O")
P           ("p"|"P")
Q           ("q"|"Q")
R           ("r"|"R")
S           ("s"|"S")
T           ("t"|"T")
U           ("u"|"U")
V           ("v"|"V")
W           ("w"|"W")
X           ("x"|"X")
Y           ("y"|"Y")
Z           ("z"|"Z")
DIGIT       [0-9]
DOT			"."
IDENTIFIER  ({LETTER}|"_")({LETTER}|{DIGIT}|"_")*
INTEGER     {DIGIT}+
FRACTION    {DOT}{INTEGER}
EXPONENT    ([eE][+-]?{INTEGER})
REAL        "-"?{INTEGER}({FRACTION}{EXPONENT}?|{EXPONENT})|{FRACTION}{EXPONENT}?
NEWLINE     "\n"

%%


{SPACE}                                ;
{NEWLINE}                              ;
"#"                                    {
                                           while (input() != '\n');
                                       }
"\""									{
											read_string();
											return STRING;
										};
{REAL}                                 {
                                           lex_float=atof(yytext);
                                           return(FLOAT);
                                       };
"-"?{INTEGER}                          {
                                           lex_int=atoi(yytext);
                                           return(INT);
                                       };
{L}{O}{O}{K}{F}{R}{O}{M}               return(LOOKFROM);
{L}{O}{O}{K}{A}{T}                     return(LOOKAT);
{L}{O}{O}{K}{U}{P}                     return(LOOKUP);
{E}{Y}{E}{D}{I}{S}{T}                  return(EYEDIST);
{W}{I}{N}{D}{O}{W}                     return(WINDOW);
{S}{C}{R}{E}{E}{N}                     return(SCREEN);
{A}{M}{B}{I}{E}{N}{T}                  return(AMBIENT);
{L}{I}{G}{H}{T}                        return(LIGHT);
{P}{O}{S}{I}{T}{I}{O}{N}               return(POSITION);
{C}{O}{L}{O}{U}{R}                     return(COLOUR);
{D}{I}{F}{F}{E}{R}{E}{N}{C}{E}         return(DIFFERENCE);
{U}{N}{I}{O}{N}                        return(UNION);
{I}{N}{T}{E}{R}{S}{E}{C}{T}{I}{O}{N}   return(INTERSECTION);
{S}{P}{H}{E}{R}{E}                     return(SPHERE);
{C}{U}{B}{E}                           return(CUBE);
{C}{Y}{L}{I}{N}{D}{E}{R}               return(CYLINDER);
{C}{O}{N}{E}                           return(CONE);
{P}{L}{A}{N}{E}                        return(PLANE);
{S}{Q}{U}{A}{R}{E}                     return(SQUARE);
{T}{O}{R}{U}{S}                        return(TORUS);
{R}{A}{D}{I}{U}{S}                     return(RADIUS);
{S}{C}{A}{L}{E}                        return(SCALE);
{R}{O}{T}{A}{T}{E}                     return(ROTATE);
{T}{R}{A}{N}{S}{L}{A}{T}{E}            return(TRANSLATE);
{T}{R}{A}{N}{S}{F}{O}{R}{M}            return(TRANSFORM);
{D}{I}{F}{F}{U}{S}{E}                  return(DIFFUSE);
{S}{P}{E}{C}{U}{L}{A}{R}               return(SPECULAR);
{T}{R}{A}{N}{S}{P}{A}{R}{E}{N}{C}{Y}   return(TRANSPARENCY);
{R}{E}{F}{L}{E}{C}{T}                  return(REFLECT);
{R}{E}{F}{R}{A}{C}{T}                  return(REFRACT);
{S}{E}{T}                              return(SET);
{P}{I}{X}{M}{A}{P}                     return(PIXMAP);
{E}{N}{D}                              return(END);
{X}                                    return(X_AXIS);
{Y}                                    return(Y_AXIS);
{Z}                                    return(Z_AXIS);
{N}{A}{M}{E}                           return(NAME);
{I}{N}{S}{T}{A}{N}{C}{E}               return(INSTANCE);

%%

void read_string()
{
	int ch;
	int max, i;

	max = 24;
	lex_string = malloc(max);
	i = 0;

	while (((ch=input()) != '"') && (ch != 0))
	{
		if (i == max)
		{
			max += 24;
			lex_string = realloc(lex_string, max);
		}

		lex_string[i++] = ch;
	}

	lex_string[i] = '\0';
}
